﻿using Microsoft.Samples.Eventing.Interop;
using System;
using System.Collections.Generic;
using System.Linq;
using System.Runtime.CompilerServices;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading.Tasks;

namespace Microsoft.Samples.Eventing.Interop
{
    public static class EtwInterop
    {
        static SortedDictionary<string, Guid> _registeredProviders = null;

        public static SortedDictionary<string, Guid> RegisteredProviders
        {
            get
            {
                if (_registeredProviders == null)
                {
                    _registeredProviders = GetETWProviders();
                }

                return _registeredProviders;
            }
        }

        public static int ProcessTrace(ulong traceHandle)
        {
            ulong[] handles = { traceHandle };
            int result = NativeMethods.ProcessTrace(handles, (uint)handles.Length, IntPtr.Zero, IntPtr.Zero);
            return result;
        }

        public static List<EventTraceProperitesManaged> QueryAllSessions()
        {
            List<EventTraceProperitesManaged> list = new List<EventTraceProperitesManaged>();

            unsafe
            {
                IntPtr ptrBuf = Marshal.AllocHGlobal(EventTraceProperties.MaxSessionBufferSize);

                try
                {
                    IntPtr[] arrayBuf = new IntPtr[NativeMethods.MAX_SESSIONS];

                    for (int i = 0; i < NativeMethods.MAX_SESSIONS; i++)
                    {
                        EventTraceProperties* pProp = (EventTraceProperties*)(((byte*)ptrBuf.ToPointer()) + EventTraceProperties.RecordSize * i);

                        IntPtr elemPtr = new IntPtr(pProp);
                        arrayBuf[i] = elemPtr;

                        pProp->Initialize();
                    }

                    GCHandle gcHandle = GCHandle.Alloc(arrayBuf, GCHandleType.Pinned);

                    try
                    {
                        IntPtr elem0 = gcHandle.AddrOfPinnedObject();

                        uint activeSessionCount = 0;
                        int status = NativeMethods.QueryAllTraces(elem0, NativeMethods.MAX_SESSIONS, ref activeSessionCount);
                        if (status == 0)
                        {
                            for (int i = 0; i < activeSessionCount; i++)
                            {
                                EventTraceProperties* pProp = (EventTraceProperties*)(((byte*)ptrBuf.ToPointer()) + EventTraceProperties.RecordSize * i);
                                list.Add(pProp->ReadAsManaged());
                            }
                        }
                    }
                    finally
                    {
                        gcHandle.Free();
                    }
                }
                finally
                {
                    Marshal.FreeHGlobal(ptrBuf);
                }
            }

            return list;
        }

        public const uint EVENT_TRACE_CONTROL_QUERY = 0;
        public const uint EVENT_TRACE_CONTROL_STOP = 1;
        public const uint EVENT_TRACE_CONTROL_UPDATE = 2;

        public static void CloseActiveSession(string sessionName)
        {
            if (IsSessionActive(sessionName, out _) == true)
            {
                EventTraceProperties prop = new EventTraceProperties();
                prop.Initialize();

                NativeMethods.ControlTrace(0, sessionName, ref prop, EVENT_TRACE_CONTROL_STOP);
            }
        }

        public static bool IsSessionActive(string sessionName, out EventTraceProperitesManaged propManaged)
        {
            EventTraceProperties prop = new EventTraceProperties();
            prop.Initialize();

            int status = NativeMethods.QueryTrace(0, sessionName, ref prop);

            if (status == 0)
            {
                propManaged = prop.ReadAsManaged();
                return true;
            }
            else if (status == EventTraceProperties.ERROR_WMI_INSTANCE_NOT_FOUND)
            {
                propManaged = new EventTraceProperitesManaged();
                // The instance name passed was not recognized as valid by a WMI data provider.
                return false;
            }

            throw new System.ComponentModel.Win32Exception(status);
        }

        static unsafe SortedDictionary<string, Guid> GetETWProviders()
        {
            var provider = new SortedDictionary<string, Guid>(StringComparer.OrdinalIgnoreCase);

            int buffSize = 0;
            var hr = NativeMethods.TdhEnumerateProviders(null, ref buffSize);
            var buffer = stackalloc byte[buffSize];
            var providersDesc = (PROVIDER_ENUMERATION_INFO*)buffer;

            hr = NativeMethods.TdhEnumerateProviders(providersDesc, ref buffSize);
            if (hr != 0)
            {
                providersDesc->NumberOfProviders = 0;
            }

            var providers = (TRACE_PROVIDER_INFO*)&providersDesc[1];
            for (int i = 0; i < providersDesc->NumberOfProviders; i++)
            {
                var name = new string((char*)&buffer[providers[i].ProviderNameOffset]);
                provider[name] = providers[i].ProviderGuid;
            }

            return provider;
        }
    }
}
