﻿using Microsoft.Samples.Eventing;
using Microsoft.Samples.Eventing.Interop;
using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Linq;
using System.Runtime.InteropServices;
using System.Text;
using System.Threading;
using System.Threading.Tasks;

namespace ConsoleApp1
{
    class Program
    {
        // .NET Console Application Exit Event
        // ; https://stackoverflow.com/questions/1119841/net-console-application-exit-event

        [DllImport("Kernel32")]
        private static extern bool SetConsoleCtrlHandler(EventHandler handler, bool add);

        private delegate bool EventHandler(CtrlType sig);

        enum CtrlType
        {
            CTRL_C_EVENT = 0,
            CTRL_BREAK_EVENT = 1,
            CTRL_CLOSE_EVENT = 2,
            CTRL_LOGOFF_EVENT = 5,
            CTRL_SHUTDOWN_EVENT = 6
        }

        static bool _disposed = false;

        static void Main(string[] _)
        {
            //if (false)
            //{
            //    PrintProviders();

            //    Console.WriteLine();
            //    Console.WriteLine();

            //    foreach (var item in EtwInterop.QueryAllSessions())
            //    {
            //        Console.WriteLine(item);
            //    }
            //}

            string sessionName = NativeMethods.KERNEL_LOGGER_NAME;

            SetConsoleCtrlHandler((sig) =>
            {
                if (_disposed == false)
                {
                    _disposed = true;
                    EtwInterop.CloseActiveSession(sessionName);
                }

                return true;
            }, true);

            int result = 0;

            ulong traceHandle = 0;
            ulong sessionHandle = 0;

            try
            {
                if (EtwInterop.IsSessionActive(sessionName, out var _) == true)
                {
                }
                else
                {
                    EventTraceProperties prop = new EventTraceProperties(true, sessionName, NativeMethods.SystemTraceControlGuid);
                    prop.SetFlags(NativeMethods.EVENT_TRACE_FLAG_PROCESS);

                    result = NativeMethods.StartTrace(out sessionHandle, sessionName, ref prop);
                    Console.WriteLine(result);
                }

                if (result == 0)
                {
                    EventTraceLogfile logFile = new EventTraceLogfile();
                    logFile.LoggerName = sessionName;
                    logFile.EventRecordCallback = EventRecordCallback;

                    logFile.ProcessTraceMode = NativeMethods.PROCESS_TRACE_MODE_EVENT_RECORD | NativeMethods.PROCESS_TRACE_MODE_REAL_TIME
                        | NativeMethods.PROCESS_TRACE_MODE_RAW_TIMESTAMP;
                    traceHandle = NativeMethods.OpenTrace(ref logFile);
                }

                if (traceHandle != 0)
                {
                    new Thread((ThreadStart)(
                    () =>
                    {
                        EtwInterop.ProcessTrace(traceHandle);
                    })).Start();

                    Console.WriteLine("Press ENTER key to exit...");
                    Console.ReadLine();
                }

                NativeMethods.CloseTrace(traceHandle);
            }
            finally
            {
                _disposed = true;
                EtwInterop.CloseActiveSession(sessionName);
            }
        }

        static private void EventRecordCallback([In] ref EventRecord eventRecord)
        {
            Console.WriteLine(DateTime.Now + ": " + eventRecord.EventHeader.ProcessId);
        }

        static void PrintProviders()
        {
            foreach (var item in EtwInterop.RegisteredProviders)
            {
                Console.WriteLine(item.Key + ": " + item.Value);
            }
        }
    }
}
